/*
 * HPT371 UDMA/ATA133 Controller Device Driver for Linux
 * Copyright (C) 2000-2003 HighPoint Technologies, Inc.
 */
#include "global.h"
#include "oslinux.h"

#if DBG
#ifdef DEBUG_LEVEL
int hpt_dbg_level = DEBUG_LEVEL;
#else
int hpt_dbg_level = 0;
#endif
#endif

#ifdef MODULE
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
		char kernel_version[] = UTS_RELEASE;
		MODULE_AUTHOR ("HighPoint Technologies, Inc.");
		MODULE_DESCRIPTION ("HPT371 Controller driver");
	#endif
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
#include <linux/reboot.h>
#endif

#ifndef DECLARE_MUTEX_LOCKED
#define DECLARE_MUTEX_LOCKED(sem) struct semaphore sem = MUTEX_LOCKED
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#define GLOBAL_DATA_ALLOC(size) ({ void *p=kmalloc(size, GFP_ATOMIC|GFP_DMA); if (p) memset(p, 0, size); p; })
#define GLOBAL_DATA_FREE(p, size) kfree(p)
#else
#define GLOBAL_DATA_ALLOC(size) scsi_init_malloc(size, GFP_ATOMIC|GFP_DMA)
#define GLOBAL_DATA_FREE(p, size) scsi_init_free((char*)p, size)
#endif

#if LINUX_VERSION_CODE == KERNEL_VERSION(2,4,2)
#define VMALLOC(size) kmalloc(size, GFP_ATOMIC)
#define VFREE(p) kfree(p)
#else
#define VMALLOC(size) vmalloc(size)
#define VFREE(p) vfree(p)
#endif

#include "glbdata.c"

typedef struct st_HPT_DPC {
	void (*dpc)(void *, UCHAR);
	void *arg;
	UCHAR flags;
} ST_HPT_DPC;

#define MAX_DPC 16
UCHAR DPC_Request_Nums = 0; 

ULONG hpt_page_offset = 0;
volatile struct Scsi_Host *hpt_vhost = 0;
static DWORD hpt_irqmasks = 0;

struct semaphore *hpt_notify_wt = 0;
struct semaphore * hpt_queue_ready = 0;
struct semaphore * hpt_queue_wt = 0;
struct task_struct *hpt_worker_thread_handle = 0; // our kernel thread
spinlock_t driver_lock = SPIN_LOCK_UNLOCKED;

PCommand pCommandBlocks=0;
PUCHAR SglBuffer[MAX_CHANNELS] = {0, };

volatile int stop_thread = 0;

void hpt_worker_thread(void *arg);
int hpt3xx_ProcInfo(char *buffer,char **start,off_t offset,int length, int hostno,int inout);
void OsSendCommand (Scsi_Cmnd *SCpnt);
void fOsCommandDone(PCommand pCmd);
static void SetInquiryData(PINQUIRYDATA inquiryData, PVDevice pVDev);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
/* notifier block to get a notify on system shutdown/halt/reboot */
static int hpt_halt(struct notifier_block *nb, ulong event, void *buf);
static struct notifier_block hpt_notifier = {
    hpt_halt, NULL, 0
};
#endif

static ULONG LastStamp = 0;
ULONG GetStamp(void)
{
	ULONG stamp;
	stamp =	jiffies;
	if(stamp == LastStamp)
		stamp++;
	LastStamp = stamp;
	return LastStamp;
}

/****************************************************************
 *	Name:	Irq_Handler
 *	Description: interrupt processing
 ****************************************************************/
static void Irq_Handler (int irq, void *dev_id, struct pt_regs *regs)
{
    unsigned long flags, f2;
	UINT  i;
	PVBus pVBus;
	BOOLEAN ret = 0;

	KdPrintE(("----- Entering Isr() -----\n"));
    spin_lock_irqsave (&io_request_lock, flags);
    spin_lock_irqsave(&driver_lock, f2);

	for(pVBus = gVBus; pVBus < &gVBus[MAX_VBUS]; pVBus++) {
		for(i = 0; i < pVBus->nInstances; i++){
		   if(HOSTDATA(shpnt)->irq[i] != irq) continue;
		   ret |= fChipInterrupt(pVBus->pChipInstance[i]);
		}
	}
	if(ret) CheckPendingCall();

	spin_unlock_irqrestore(&driver_lock, f2);
    spin_unlock_irqrestore (&io_request_lock, flags);
	KdPrintE(("----- Leaving Isr() -----\n"));
}

/* call this function inside driver_lock */
void Check_Idle_Call(void)
{
    if (mWaitingForIdle(gVBus)) {
		CheckIdleCall();
        up(hpt_queue_wt);
	}
}

/////////////////////////////////////////////
void Queue_SC(Scsi_Cmnd **SC_Q,Scsi_Cmnd *SCpnt)
{
	if(*SC_Q==NULL)
		SCpnt->SCp.ptr = (char *)SCpnt;
	else
	{
		SCpnt->SCp.ptr = (*SC_Q)->SCp.ptr;
		(*SC_Q)->SCp.ptr = (char *)SCpnt;
	}
	*SC_Q = SCpnt;
}

void Release_SC(Scsi_Cmnd **SC_Q,Scsi_Cmnd *SCpnt)
{
	Scsi_Cmnd *TmpSCpnt;
    TmpSCpnt = *SC_Q;
	
	if(SCpnt->SCp.ptr == (char *)SCpnt)/*it means SCpnt is the last one in CURRCMDs*/
		*SC_Q = NULL;
	else
	{
		while(TmpSCpnt->SCp.ptr != (char *)SCpnt){
			TmpSCpnt = (Scsi_Cmnd *)TmpSCpnt->SCp.ptr;
		}
		TmpSCpnt->SCp.ptr = SCpnt->SCp.ptr;
		if(*SC_Q == SCpnt)
			*SC_Q = TmpSCpnt;
	}
}
void ScsiDone(Scsi_Cmnd *SCpnt)
{
	SCpnt->scsi_done(SCpnt);
	Release_SC(&CURRENT_Q,SCpnt);
    CURRCMDs--;

	if (CURRCMDs == 0)
	{
		if(DPC_Request_Nums == 0)
		    Check_Idle_Call();
/*check pending_Q*/
        while(PENDING_Q != NULL)
		{
			SCpnt = (Scsi_Cmnd *)PENDING_Q->SCp.ptr;
            Release_SC(&PENDING_Q,SCpnt);
			CallAfterReturn((DPC_PROC)OsSendCommand, SCpnt);
		}
	}
}


/****************************************************************
 *	Name:	OsSendCommand
 *	Description:	execute a queued command from current_Q.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		viod
 ****************************************************************/
void OsSendCommand (Scsi_Cmnd *SCpnt)
{
	PCommand pCmd;
	PVDevice pVDev = gVBus[SCpnt->channel].pVDevice[SCpnt->target];

    Queue_SC(&CURRENT_Q,SCpnt);
	CURRCMDs++;

	switch(SCpnt->cmnd[0]) {
	case START_STOP:
		SCpnt->result = (DID_OK<<16);
        break;

	case INQUIRY:
		ZeroMemory(SCpnt->request_buffer, SCpnt->request_bufflen);
		SetInquiryData(SCpnt->request_buffer, pVDev);
		SCpnt->result = (DID_OK<<16);
        break;

	case READ_CAPACITY:
    {
		UCHAR swip[4];
		//
		// Claim 512 byte blocks (big-endian).
		//				
		((PREAD_CAPACITY_DATA)SCpnt->request_buffer)->BytesPerBlock = 0x20000;
		*(ULONG*)swip = pVDev->VDeviceCapacity - 1;
		((PREAD_CAPACITY_DATA)SCpnt->request_buffer)->LogicalBlockAddress =
			((ULONG)swip[0] << 24) |  ((ULONG)swip[1] << 16) | ((ULONG)swip[2] << 8) | (ULONG)swip[3];
		SCpnt->result = (DID_OK<<16);
        break;
	}

	case READ_6:
	case WRITE_6:
	case 0x13:      /*SCSIOP_VERIFY6*/
	case READ_10:
	case WRITE_10:
	case VERIFY:
		{						
			pCmd = AllocateCommand();
			HPT_ASSERT(pCmd);

			switch (SCpnt->cmnd[0]) {
			case READ_6:
			case WRITE_6:
			case 0x13:      /*SCSIOP_VERIFY6*/
				pCmd->uCmd.Ide.Lba = ((ULONG)SCpnt->cmnd[1] << 16) | ((ULONG)SCpnt->cmnd[2] << 8) | (ULONG)SCpnt->cmnd[3];
				pCmd->uCmd.Ide.nSectors = SCpnt->cmnd[4];
				break;
			default:
			    pCmd->uCmd.Ide.Lba = (ULONG)SCpnt->cmnd[5] | ((ULONG)SCpnt->cmnd[4] << 8) |
			         ((ULONG)SCpnt->cmnd[3] << 16) | ((ULONG)SCpnt->cmnd[2] << 24);
			    pCmd->uCmd.Ide.nSectors = (USHORT)SCpnt->cmnd[8] | ((USHORT)SCpnt->cmnd[7]<<8);
			    break;
			}
		    switch (SCpnt->cmnd[0]) {
		    case READ_10:
			case READ_6:
		    	pCmd->uCmd.Ide.Command = IDE_COMMAND_READ;
			    pCmd->cf_data_in = 1;
		    	break;
		    case WRITE_10:
			case WRITE_6:
		    	pCmd->uCmd.Ide.Command = IDE_COMMAND_WRITE;
		    	pCmd->cf_data_out = 1;
		    	break;
		    case VERIFY:
		    case 0x13:      /*SCSIOP_VERIFY6*/
		    	pCmd->uCmd.Ide.Command = IDE_COMMAND_VERIFY;
		    	break;
		    }
			KdPrint(("//============\n"));
            KdPrintI(("Os send Command: SCpnt= 0x%x, pVDev=0x%x cmd=%x nSectors=%d lba=%lx\n",SCpnt, pVDev, pCmd->uCmd.Ide.Command, pCmd->uCmd.Ide.nSectors, pCmd->uCmd.Ide.Lba));
            pCmd->pOrgCommand = SCpnt;
			pCmd->pVDevice = pVDev;
			pCmd->pfnBuildSgl = fOsBuildSgl;
			pCmd->pfnCompletion = fOsCommandDone;
		    pVDev->pfnSendCommand(pCmd);
		    goto out;
		}

	case TEST_UNIT_READY:
		SCpnt->result = (DID_OK<<16);
        break;

	default:
		SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
        break;
    }
	ScsiDone(SCpnt);
out:
	CheckPendingCall();
}

/****************************************************************
 *	Name:	hpt3xx_QueueCommand
 *	Description:	Process a queued command from the SCSI manager.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *					done  - Pointer to done function to call.
 *	Returns:		Status code.
 ****************************************************************/
int hpt3xx_QueueCommand (Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *))
{
    unsigned long flags;
    PVDevice pVDev;

    KdPrintE(("\nhpt3xx_QueueCommand() %d/%d/%d cdb=(%x-%x-%x)\n",
        SCpnt->channel, SCpnt->target, SCpnt->lun,
	    	*((PULONG)&SCpnt->cmnd), *((PULONG)&SCpnt->cmnd+1), *((PULONG)&SCpnt->cmnd+2)));
    SCpnt->scsi_done = (void *)done;

    /*
     * start the command. Use spinlock to avoid sending more than one 
     * command at same time
     */
	spin_lock_irqsave(&driver_lock, flags);

	if (SCpnt->channel>=MAX_VBUS || SCpnt->target>=MAX_DEVICES || SCpnt->lun)
		pVDev = 0;
	else
		pVDev = gVBus[SCpnt->channel].pVDevice[SCpnt->target];
	if (!pVDev) {
		SCpnt->result = (DID_BAD_TARGET<<16);
		SCpnt->scsi_done(SCpnt);
		goto out;
	}
    if(mWaitingForIdle(gVBus))
	{
		Queue_SC(&PENDING_Q,SCpnt);
		goto out;
	}

    if ((CURRCMDs == 0) && (DPC_Request_Nums == 0))
		Check_Idle_Call();

    OsSendCommand(SCpnt);
out:
	spin_unlock_irqrestore (&driver_lock, flags);
	KdPrintE(("----- Leaving hpt3xx_queueCommand() -----\n"));
	return 0;
}

/****************************************************************
 *	Name:	internal_done
 *	Description:	Done handler for non-queued commands
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		Nothing.
 ****************************************************************/
static void internal_done (Scsi_Cmnd * SCpnt)
{
	SCpnt->SCp.Status++;
}

/****************************************************************
 *	Name:	hpt3xx_Command
 *	Description:	Process a command from the SCSI manager.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		Status code.
 ****************************************************************/
int hpt3xx_Command (Scsi_Cmnd *SCpnt)
{
	DWORD timeout;
	
    SCpnt->SCp.Status = 0;
	hpt3xx_QueueCommand (SCpnt, internal_done);

	timeout = jiffies + 60 * HZ; // should be longer than hard-reset time
	while (!SCpnt->SCp.Status && time_before(jiffies, timeout)) schedule();

	if (!SCpnt->SCp.Status) SCpnt->result = (DID_ERROR<<16);
	return SCpnt->result;
}

/****************************************************************
 *	Name:	hpt3xx_Detect
 *	Description:	Detect and initialize our boards.
 *	Parameters:		tpnt - Pointer to SCSI host template structure.
 *	Returns:		Number of adapters installed.
 ****************************************************************/
int hpt3xx_Detect (Scsi_Host_Template *tpnt)
{
	PChipInstance pChip;
	PChannel  pChan;
	struct pci_dev *pPciDev;
	int i,irqnum;

	printk("%s\nVersion %s, Compiled %s %s\n", 
		DRIVER_NAME, DRIVER_VERSION,  __DATE__, __TIME__);
	/* 
	 * calculate page_offset.
	 */
	hpt_page_offset = (ULONG)tpnt - virt_to_phys(tpnt);/*0xc0000000*/

	if (!pci_present()) {
		printk(KERN_ERR "PCI BIOS not present\n");
		return 0;
	}
	
	/* setup command blocks */
	pCommandBlocks = VMALLOC(sizeof(struct _Command) * MAX_COMMAND_BLOCKS_FOR_EACH_VBUS);
	if (!pCommandBlocks) {
		hpt_printk(("insufficient memory\n"));
		return 0;
	}
	for (i=0; i<MAX_COMMAND_BLOCKS_FOR_EACH_VBUS; i++)
		FreeCommand(&pCommandBlocks[i]);
	
	if (Initialize()==0) {
		if(gVBus[0].nInstances == 0)
		    return 0;
	}

	//
	// Allocate a Noncached Extension to use for scatter/gather list
	//
	for (pChan=pChanStart; pChan<pChanEnd; pChan++) {
		i = pChan-pChanStart;
        SglBuffer[i] = GLOBAL_DATA_ALLOC(MAX_SG_DESCRIPTORS * sizeof(SCAT_GATH) + 16);
		if (!SglBuffer[i]) {
			hpt_printk(("can't alloc SG buffer, exit\n"));
			vfree(pCommandBlocks);
			return 0;
		}
		pChan->SgPhysicalAddr = fOsPhysicalAddress(SglBuffer[i] + 15) & 0xfffffff0;
		pChan->pSgTable = (FPSCAT_GATH)((void *)(hpt_page_offset | pChan->SgPhysicalAddr));
	}

	/*
	 * register a virtual host.
	 */
	hpt_vhost = scsi_register(tpnt, sizeof(struct hpt3xx_HOSTDATA));
	hpt_vhost->io_port		= 0;
	hpt_vhost->n_io_port	= 0;
	hpt_vhost->max_channel	= 0;  /* channel <= max_channel */
	hpt_vhost->max_id		= MAX_DEVICES;  /* id < max_id */
	hpt_vhost->max_lun      = 1;  /* lun < max_lun */

	PENDING_Q	            = 0;
	CURRENT_Q	            = 0;
	CURRCMDs		        = 0;
	HOSTDATA(shpnt)->status = FALSE;

	/* 
	 * search for HPT371 chip 
	 */
	pPciDev = NULL;
	i = 0;
	while ((pPciDev=pci_find_device(HPT_VENDORID, HPT371_DEVICEID, pPciDev))){
		if (i >= MAX_CHIPS) break;
		for(pChip = gChipInstances; pChip < pChipEnd; pChip++)
			if((pChip->pci_bus == pPciDev->bus->number) && 
				(((pChip->pci_dev<<3) | pChip->pci_func) == pPciDev->devfn))
				break;
		HOSTDATA(shpnt)->irq[pChip-gChipInstances] = (UCHAR)pPciDev->irq;
		i++;
	}

	/*
	 * set up irq handler
	 */	
	for(pChip = gChipInstances; pChip < pChipEnd; pChip++) {

		irqnum = HOSTDATA(shpnt)->irq[pChip-gChipInstances];
		if ((hpt_irqmasks & (1<<irqnum))==0) {
			hpt_irqmasks |= (1<<irqnum);
			if (request_irq(irqnum, Irq_Handler, SA_SHIRQ, PROC_DIR_NAME, gChipInstances) < 0) {
				if (request_irq(irqnum, Irq_Handler, SA_INTERRUPT | SA_SHIRQ, PROC_DIR_NAME, gChipInstances) < 0 ){
					hpt_printk (("Unable to allocate IRQ\n"));
					return -1;
				}
			}
		}
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
    register_reboot_notifier(&hpt_notifier);
#endif

	return 1;
}

/****************************************************************
 *	Name:	hpt3xx_Abort
 *	Description:	Process the Abort command from the SCSI manager.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		Always snooze.
 ****************************************************************/
int hpt3xx_Abort (Scsi_Cmnd *SCpnt)
{
	hpt_printk(("Abort called pid: %ld target: %x lun: %x flags: %x"
			   " reason %x\n", SCpnt->pid, SCpnt->target, SCpnt->lun, SCpnt->flags, 
			   SCpnt->abort_reason));
	return FAILED;
}

/**************************************************************************
 *    The reset function will reset the SCSI bus and abort any executing
 *    commands. 
 *
 *	Parameters:	   SCpnt - Pointer to SCSI command structure.
 *				   flags - Flags about the reset command
 *  Returns:       DID_RESET in cmd.host_byte of aborted command(s) 
 *  Note:          Resetting the bus always succeeds - is has to, otherwise the
 *                 kernel will panic! Try a surgical technique - sending a BUS
 *                 DEVICE RESET message - on the offending target before pulling
 *                 the SCSI bus reset line.
 **************************************************************************/
int hpt3xx_Reset (Scsi_Cmnd *SCpnt)
{
	/* The Scsi_Cmnd  is passed so that the reset routine can figure
     * out which host adapter should be reset, and also which command
     * within the command block was responsible for the reset in
     * the first place.
	 * Only one virtual host in hpt3xx driver -----hpt_vhost
	 */
	int Found = 0;
	Scsi_Cmnd *SCtmp;
	unsigned long flags;
	DWORD timeout;
	
	if (SCpnt==NULL) {
		printk(KERN_WARNING "(scsi?:?:?:?) Reset called with NULL Scsi_Cmnd "
		  "pointer, failing.\n");
		return FAILED;
	}

	hpt_printk(("Reset called pid: %ld target: %x lun: %x flags: %x\n",
		SCpnt->pid, SCpnt->target, SCpnt->lun, SCpnt->flags));
	
	/*
	 * Return immediately if reset is in progress.
	 */
	if (HOSTDATA(shpnt)->status == TRUE) {
		hpt_printk(("reset: exit, already in reset.\n"));
		return FAILED;
	}

	spin_lock_irqsave(&driver_lock,flags);

	HOSTDATA(shpnt)->status = TRUE;
	timeout = jiffies + 60 * HZ;	
	while (DPC_Request_Nums && time_before(jiffies, timeout)) {
		spin_unlock_irqrestore(&driver_lock,flags);
		schedule();		
		spin_lock_irqsave(&driver_lock,flags);
	}
	
	fResetVBus();
	
	HPT_ASSERT(PENDING_Q==NULL);

   /*
    * Wake-up all awaiting commands with DID_RESET.
    */
	while(CURRENT_Q != NULL)
	{
		SCtmp = (Scsi_Cmnd *)CURRENT_Q->SCp.ptr;
		Release_SC(&CURRENT_Q,SCtmp);
		CURRCMDs--;
		SCtmp->result = (DID_RESET << 16);
		SCtmp->scsi_done(SCtmp);
		if(SCtmp == SCpnt)
			Found = 1;
	}
	/*
     * If the involved command was not in a driver queue, and the 
     * scsi driver told us reset is synchronous, and the command is not 
     * currently in the waiting list, complete it with DID_RESET status,
     * in order to keep it alive.
     */
	if(!Found)
	{
		SCpnt->result = DID_RESET << 16;
		SCpnt->scsi_done(SCpnt);
	}

	/*
	 * Now tell the mid-level code what we did here.  Since
	 * we have restarted all of the outstanding commands,
	 * then report SUCCESS.
	 */
	HOSTDATA(shpnt)->status = FALSE;
	up(hpt_queue_wt);
	spin_unlock_irqrestore(&driver_lock,flags);
	return SUCCESS;
}

/****************************************************************
 *	Name:	hpt3xx_cleanup
 *	Description:	cleanup routine
 ****************************************************************/
static void hpt3xx_cleanup(void)
{
	PVBus pVBus;
	PVDevice pVDev;
	int i;
	/* flush all drives */
	FOR_EACH_DEV_ON_ALL_VBUS(pVBus, pVDev, i) {
		fFlushVDev(pVDev);
	}
}

/****************************************************************
 *	Name:	hpt3xx_Release
 *	Description:	Release resources allocated for a single each adapter.
 *	Parameters:		pshost - Pointer to SCSI command structure.
 *	Returns:		zero.
 ****************************************************************/
int hpt3xx_Release(struct Scsi_Host *pshost)
{
	int irq;
	PChannel pChan;

	vfree(pCommandBlocks);

	/* release IRQ resource */
	KdPrintI(("\n release resources allocated.\n"));
	for (irq=0; irq<32; irq++) {
		if ((1<<irq) &hpt_irqmasks)
			free_irq (irq, gChipInstances);
	}

	for (pChan=pChanStart; pChan<pChanEnd; pChan++)
		GLOBAL_DATA_FREE(SglBuffer[pChan-pChanStart], MAX_SG_DESCRIPTORS * sizeof(SCAT_GATH)+16);

    scsi_unregister(pshost);

    return 0;
}

/****************************************************************
 *	Name:	hpt3xx_BiosParam
 *	Description:	Process the biosparam request from the SCSI manager to
 *					return C/H/S data.
 *	Parameters:		disk - Pointer to SCSI disk structure.
 *					dev	 - Major/minor number from kernel.
 *					geom - Pointer to integer array to place geometry data.
 *	Returns:		zero.
 ****************************************************************/
int hpt3xx_BiosParam (Scsi_Disk *disk, kdev_t dev, int geom[])
{
	int heads, sectors, cylinders;
	
	/* Default heads (64) & sectors (32) */
	/* Handle extended translation size for logical drives > 1Gb */
	if (disk->capacity >= 0x200000) {
		heads = 255;
		sectors = 63;
	}
	else {
		heads = 64;
		sectors = 32;
	}
	cylinders = disk->capacity / (heads * sectors);
	
	/* return result */
	geom[0] = heads;
	geom[1] = sectors;
	geom[2] = cylinders;
	
	return 0;
}

/****************************************************************
 *	Name:	hpt_halt
 *	Description:	shutdown routine
 ****************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
static int hpt_halt(struct notifier_block *nb, ulong event, void *buf)
{
	if (event != SYS_RESTART && event != SYS_HALT && event != SYS_POWER_OFF)
		return NOTIFY_DONE;
		
	hpt3xx_cleanup();
		
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,22)
	unregister_reboot_notifier(&hpt_notifier);
#endif
    return NOTIFY_OK;
}
#endif

int fOsBuildSgl(PCommand pCmd, FPSCAT_GATH pSg, int logical)
{
	Scsi_Cmnd *SCpnt = (Scsi_Cmnd *)pCmd->pOrgCommand;
	int       idx = 0;
	ADDRESS   dataPointer = 0;
	ULONG     bytesLeft   = SCpnt->request_bufflen;
	ULONG     length;
	PUCHAR    p = (PUCHAR)SCpnt->request_buffer;
	struct scatterlist *sgList = (struct scatterlist *)SCpnt->request_buffer;

	if (logical) {
		if(SCpnt->use_sg){
			for (idx = 0; idx < SCpnt->use_sg; idx++) {
				pSg[idx].dSgAddress = (ULONG)sgList[idx].address;
				pSg[idx].wSgSize = sgList[idx].length;
				pSg[idx].wSgFlag = (idx==SCpnt->use_sg-1)? SG_FLAG_EOT : 0;
			}
		}
		else{
			pSg->dSgAddress = (ULONG)SCpnt->request_buffer;
			pSg->wSgSize = (USHORT)SCpnt->request_bufflen;
			pSg->wSgFlag = SG_FLAG_EOT;
		}
	}
	else 
	{
		if(SCpnt->use_sg)
		{
			for (idx = 0; idx < SCpnt->use_sg; idx++) {
				pSg[idx].dSgAddress = fOsPhysicalAddress(sgList[idx].address);
				if (pSg[idx].dSgAddress & 1) return FALSE;
				pSg[idx].wSgSize = sgList[idx].length;
				if ((pSg[idx].wSgSize & 3) || pSg[idx].wSgSize<8) return FALSE;
				pSg[idx].wSgFlag = (idx==SCpnt->use_sg-1)? SG_FLAG_EOT : 0;
			}
		}
		else
		{
			dataPointer = (ADDRESS)fOsPhysicalAddress(SCpnt->request_buffer);
			KdPrint(("Build SGTable, SCpnt->request_buffer=0x%x, dataPointer=0x%x", SCpnt->request_buffer,dataPointer));
            /*
			 * the address must be word align; the size must be word align
			 */
			if (((ULONG)dataPointer & 0x1) || (bytesLeft & 0x1))
				return FALSE;
			length = 0x1000 - ((ULONG)dataPointer & 0xFFF);
			do {
				if (idx >= MAX_SG_DESCRIPTORS) return FALSE;
				pSg[idx].dSgAddress = fOsPhysicalAddress(p);
				if (length > bytesLeft) length = bytesLeft;
				pSg[idx].wSgSize = (USHORT)length;
				bytesLeft -= length;
				pSg[idx].wSgFlag = (bytesLeft)? 0 : SG_FLAG_EOT;
				p+=length;
				length = 0x1000;
				idx++;
			}
			while (bytesLeft);
			return TRUE;
		}
	}
	return TRUE;
}

void fOsCommandDone(PCommand pCmd)
{
	Scsi_Cmnd *SCpnt = (Scsi_Cmnd *)pCmd->pOrgCommand;

	switch(pCmd->Result) {
	case RETURN_SUCCESS:
		KdPrint(("RETURN_SUCCESS\n"));
		SCpnt->result = (DID_OK<<16);
		break;
	case RETURN_BAD_DEVICE:
		KdPrint(("RETURN_BAD_DEVICE"));
		SCpnt->result = (DID_BAD_TARGET<<16);
		break;
	case RETURN_DEVICE_BUSY:
		KdPrint(("RETURN_DEVICE_BUSY"));
		SCpnt->result = (DID_BUS_BUSY<<16);
		break;
	case RETURN_SELECTION_TIMEOUT:
		KdPrint(("RETURN_SELECTION_TIMEOUT"));
		SCpnt->result = (DID_NO_CONNECT<<16);
		break;
	case RETURN_BUS_RESET:
		KdPrint(("RETURN_BUS_RESET"));
		SCpnt->result = (DID_RESET<<16);
		break;
	default:
		KdPrint(("RETURN_ERROR"));
		SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
		break;
	}
	FreeCommand(pCmd);
	ScsiDone(SCpnt);
}

ULONG fAsMapLbaToCHS(ULONG Lba, WORD sectorXhead, BYTE sector_per_track)
{
	/* LBA = ((C * heads + H) * sectors_per_track ) + S - 1 */
	ULONG C, H, S;
	USHORT heads = sectorXhead / sector_per_track;
	S = Lba % sector_per_track + 1;
	H = Lba / sector_per_track % heads;
	C = Lba / sector_per_track / heads;
	/*
	 * return 28bit address.
	 */
	return S | C<<8 | H<<24;
}

void OsSetDeviceTable(PDevice pDevice, PIDENTIFY_DATA pIdentify)
{
    MemoryCopy(&pDevice->IdentifyData, pIdentify, sizeof(IDENTIFY_DATA2));
}

static void SetInquiryData(PINQUIRYDATA inquiryData, PVDevice pVDev)
{
	int i;		
	inquiryData->DeviceType = TYPE_DISK; /*DIRECT_ACCESS_DEVICE*/
	inquiryData->AdditionalLength = (UCHAR)(sizeof(PINQUIRYDATA) - 5);
#ifndef SERIAL_CMDS
	inquiryData->CommandQueue = 1;
#endif

	switch(pVDev->VDeviceType) {
	case VD_SINGLE_DISK:
	case VD_ATAPI:
	case VD_REMOVABLE:
		//
		// Set the removable bit, if applicable.
		//
		if ((pVDev->u.disk.df_removable_drive) || (pVDev->u.disk.IdentifyData.GeneralConfiguration & 0x80))
			inquiryData->RemovableMedia = 1;

		//
		// Fill in vendor identification fields.
		//
		for (i = 0; i < 20; i += 2)
		{				
			inquiryData->VendorId[i] 	= ((PUCHAR)pVDev->u.disk.IdentifyData.ModelNumber)[i + 1];
			inquiryData->VendorId[i+1] 	= ((PUCHAR)pVDev->u.disk.IdentifyData.ModelNumber)[i];

		}

		//
		// Initialize unused portion of product id.
		//
		for (i = 0; i < 4; i++) inquiryData->ProductId[12+i] = ' ';

		//
		// Move firmware revision from IDENTIFY data to
		// product revision in INQUIRY data.
		//
		for (i = 0; i < 4; i += 2)
		{				
			inquiryData->ProductRevisionLevel[i] 	= ((PUCHAR)pVDev->u.disk.IdentifyData.FirmwareRevision)[i+1];
			inquiryData->ProductRevisionLevel[i+1] 	= ((PUCHAR)pVDev->u.disk.IdentifyData.FirmwareRevision)[i];
		}
		break;
	default:
		memcpy(&inquiryData->VendorId, "HPT3xx  ", 8);
		memcpy(&inquiryData->ProductRevisionLevel, "3.00", 4);
	    break;
	}
}

void hpt_refresh_logical_devices(void)
{
}

void hpt_set_device_offline(int id)
{
	Scsi_Device *SDptr = hpt_vhost->host_queue;
	while (SDptr) {
		if (SDptr->id == id) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
			SDptr->online = 0;
#endif
			return;
		}
		SDptr = SDptr->next;
	}
}

static void hpt_scsi_unregister_module(int module_type, void *ptr)
{
	hpt3xx_cleanup();
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
		unregister_reboot_notifier(&hpt_notifier);
#endif
	scsi_unregister_module(module_type, ptr);
}

#ifdef START_WORKER_THREAD_BEFORE_INIT
#define init_module scsi_init_module
#endif

/* 
 * use with care: scsi_unregister_module() may vary on diffrent kernel versions
 */
#undef scsi_unregister_module
#define scsi_unregister_module hpt_scsi_unregister_module
static Scsi_Host_Template driver_template = HPT3xx;
#include "scsi_module.c"

#ifdef START_WORKER_THREAD_BEFORE_INIT
#undef init_module
int init_module(void)
{
	int ret;
	launch_worker_thread();
	ret = scsi_init_module();
	if (ret && hpt_worker_thread_handle) {
		DECLARE_MUTEX_LOCKED(sem);
		hpt_notify_wt = &sem;
		stop_thread = 1;
		up(hpt_queue_ready);
		down(&sem);
		hpt_notify_wt = NULL;
	}
	return ret;
}
#endif

/* PCI space access */
UCHAR pcicfg_read_byte (UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg)
{
	UCHAR v;
	pcibios_read_config_byte(bus, (dev<<3)|func, reg, &v);
	return v;
}
WORD pcicfg_read_word(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg)
{
	WORD v;
	pcibios_read_config_word(bus, (dev<<3)|func, reg, &v);
	return v;
}
DWORD pcicfg_read_dword(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg)
{
	DWORD v;
	pcibios_read_config_dword(bus, (dev<<3)|func, reg, (unsigned int *)&v);
	return v;
}
void pcicfg_write_byte (UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, UCHAR v)
{
	pcibios_write_config_byte(bus, (dev<<3)|func, reg, v);
}
void pcicfg_write_word(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, WORD v)
{
	pcibios_write_config_word(bus, (dev<<3)|func, reg, v);
}
void pcicfg_write_dword(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, DWORD v)
{
	pcibios_write_config_dword(bus, (dev<<3)|func, reg, v);
}

BYTE hpt_inp(unsigned port) { return inb(port); }
WORD hpt_inpw(unsigned port) { return inw(port); }
DWORD hpt_inpd(unsigned port) { return inl(port); }
void hpt_outp(unsigned port, BYTE value) { outb(value, port); }
void hpt_outpw(unsigned port, WORD value) { outw(value, port); }
void hpt_outpd(unsigned port, DWORD value) { outl(value, port); }
void hpt_repinsw(unsigned ioport, ADDRESS buffer, WORD count) { insw(ioport, buffer, count); }
void hpt_repoutsw(unsigned ioport, ADDRESS buffer, WORD count) { outsw(ioport, buffer, count); }

ULONG fOsPhysicalAddress(void *addr) { return ((ULONG)addr & ~hpt_page_offset); }
void StallExec(DWORD x) { (x < MAX_UDELAY_MS*1000)? udelay(x) : mdelay(x/1000); }

void *os_alloc_page(_VBUS_ARG0) { return VMALLOC(0x1000); }
void *os_alloc_dma_page(_VBUS_ARG0) { return kmalloc(0x1000, GFP_ATOMIC|GFP_DMA); }
void os_free_page(_VBUS_ARG void *p) { VFREE(p); }
void os_free_dma_page(_VBUS_ARG void *p) { kfree(p); }

void DoXor1(ULONG *p0, ULONG *p1, ULONG *p2, UINT nBytes)
{
    UINT i;
	for (i = 0; i < nBytes / 4; i++) *p0++ = *p1++ ^ *p2++;
}

void DoXor2(ULONG *p0, ULONG *p2, UINT nBytes)
{
    UINT i;
	for (i = 0; i < nBytes / 4; i++) *p0++ ^= *p2++;
}

int os_printk(char *fmt, ...)
{
	va_list ap;
	char buf[256];
	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	return printk(buf);
}

int os_memcmp(const void *cs, const void *ct, unsigned len) { return memcmp(cs, ct, len); }
void os_memcpy(void *to, const void *from, unsigned len) { memcpy(to, from, len); }
void os_memset(void *s, char c, unsigned len) { memset(s, c, len); }
unsigned os_strlen(const char *s) { return strlen(s); }

EXPORT_NO_SYMBOLS;
