/***************************************************************************
 * File:         init.c
 * Description:  HPT3xx driver initialization
 * Copyright (c) 2000-2003 HighPoint Technologies, Inc. All rights reserved.
 ***************************************************************************/

#include "global.h"


int Initialize(void)
{
    int nDev;
    PChipInstance pChip = gChipInstances;
	PVBus pVBus;
	PVDevice pVDev;
	int j;

    ZeroMemory(gVBus, sizeof(VBus)*MAX_VBUS);


	KdPrint(("Scan for chips..."));
    /*
     * Scan Chip
     */
    for ( ; ; ) 
	{
		ZeroMemory(pChip, sizeof(ChipInstance));
		if(GetNextPciInfo(pChip) == 0)
            break;
        pChip++;
		if (pChip-gChipInstances>=MAX_CHIPS) break;
    }

    if(pChip == gChipInstances) {
    	KdPrint(("No supported chip found"));
        return(0);
	}


    /*
     * Initial all chips and the channels attached on them
     */
    pChanStart = (PChannel)pChip;
    pDataPool = (char *)pChip;
    for(pChip = gChipInstances; pChip < pChipEnd; pChip++) 
		pChip->ftbl.pfnInitialize(pChip, INIT_ENTRY);    
	
	for(pChip = gChipInstances; pChip < pChipEnd; pChip++)
		pChip->ftbl.pfnControlIrq(pChip, 0);
		
	KdPrint(("Scan for devices..."));
    /*
     * scan all devices on each channel
     */
    nDev = 0;
    pVDevStart = (PVDevice)pDataPool;
    for(pChip = gChipInstances; pChip < pChipEnd; pChip++) {
        nDev += fDeScanAllDevices(pChip);
        /* 
		 * check ATA/133 device to adjust DPLL clock. 
		 */
		if (pChip->ftbl.pfnInit2) pChip->ftbl.pfnInit2(pChip);
    }


	/* register all logical devices on vbus */
	fRegisterVDevices();

	/* mark bootable devices */
	{
	    FOR_EACH_DEV_ON_ALL_VBUS(pVBus, pVDev, j) {
			fCheckBootable(pVDev);
		}
		/* enable interrupt */
		for(pChip = gChipInstances; pChip < pChipEnd; pChip++)
			pChip->ftbl.pfnControlIrq(pChip, 1);
	}


    return(nDev); 
}

PChannel fGetChannelTable(void)
{    
     PChannel pChan = (PChannel)pDataPool;
     pDataPool += sizeof(Channel);
     ZeroMemory((char *)pChan,sizeof(Channel));
     return(pChan);
}

PDevice  fGetDeviceTable(void)
{
    PVDevice pVDev = (PVDevice)pDataPool;
    pDataPool += sizeof(VDevice);
    ZeroMemory((char *)pVDev,sizeof(VDevice));
    return &pVDev->u.disk;
}

void  OsReturnTable(PDevice pDevice)
{
    pDataPool = (char *)Map2pVDevice(pDevice); 
}

